import os
import asyncio
import random
import datetime
from typing import List, Tuple, Dict

import backoff
from fastapi import HTTPException
from pydantic import ValidationError

import aiohttp
from aiolimiter import AsyncLimiter
from openai import AsyncStream
from openai.types.chat import ChatCompletionChunk

from logdetective.extractors import Extractor
from logdetective.utils import (
    compute_certainty,
    prompt_to_messages,
    format_snippets,
    mine_logs,
)
from logdetective.server.config import (
    LOG,
    SERVER_CONFIG,
    PROMPT_CONFIG,
    CLIENT,
)
from logdetective.server.models import (
    AnalyzedSnippet,
    InferenceConfig,
    Explanation,
    StagedResponse,
    SnippetAnalysis,
    RatedSnippetAnalysis,
    Response,
)
from logdetective.server.utils import (
    format_analyzed_snippets,
    should_we_giveup,
    we_give_up,
    filter_snippets,
    construct_final_prompt,
)

LLM_CPP_SERVER_TIMEOUT = os.environ.get("LLAMA_CPP_SERVER_TIMEOUT", 600)


# pylint: disable=too-many-positional-arguments, too-many-arguments
@backoff.on_exception(
    lambda: backoff.constant([10, 30, 120]),
    aiohttp.ClientResponseError,
    max_tries=4,  # 4 tries and 3 retries
    jitter=lambda wait_gen_value: random.uniform(wait_gen_value, wait_gen_value + 30),
    giveup=should_we_giveup,
    raise_on_giveup=False,
    on_giveup=we_give_up,
)
async def call_llm(
    messages: List[Dict[str, str]],
    inference_cfg: InferenceConfig,
    async_request_limiter: AsyncLimiter,
    stream: bool = False,
    structured_output: dict | None = None,
    index: int | None = None,
) -> Explanation:
    """Submit prompt to LLM.
    inference_cfg: The configuration section from the config.json representing
    the relevant inference server for this request.
    """
    subject = "the text" if index is None else f"snippet no. {index}"
    lengths = {m["role"]: len(m["content"]) for m in messages}
    LOG.info("Analyzing %s, Message size in chars: %s", subject, str(lengths))

    LOG.info("Submitting to /v1/chat/completions endpoint")

    kwargs = {}

    # OpenAI API does not guarantee that the behavior for parameter set to `None`
    # and parameter not given at all is the same.
    # We build a dictionary of parameters based on the configuration.
    if inference_cfg.log_probs:
        LOG.info("Requesting log probabilities from LLM")
        kwargs["logprobs"] = inference_cfg.log_probs
    if structured_output:
        LOG.info("Requesting structured output from LLM")
        response_format = {
            "type": "json_schema",
            "json_schema": {
                "name": "rated-snippet-analysis",
                "schema": structured_output,
            },
        }
        kwargs["response_format"] = response_format

    async with async_request_limiter:
        response = await CLIENT.chat.completions.create(
            messages=messages,
            max_tokens=inference_cfg.max_tokens,
            stream=stream,
            model=inference_cfg.model,
            temperature=inference_cfg.temperature,
            **kwargs,
        )

    if not response.choices[0].message.content:
        LOG.error("No response content recieved from %s", inference_cfg.url)
        raise RuntimeError()

    message_content = response.choices[0].message.content

    if response.choices[0].logprobs and response.choices[0].logprobs.content:
        logprobs = [e.to_dict() for e in response.choices[0].logprobs.content]
    else:
        logprobs = None

    return Explanation(
        text=message_content,
        logprobs=logprobs,
    )


@backoff.on_exception(
    lambda: backoff.constant([10, 30, 120]),
    aiohttp.ClientResponseError,
    max_tries=4,  # 4 tries and 3 retries
    jitter=lambda wait_gen_value: random.uniform(wait_gen_value, wait_gen_value + 30),
    giveup=should_we_giveup,
    raise_on_giveup=False,
    on_giveup=we_give_up,
)
async def call_llm_stream(
    messages: List[Dict[str, str]],
    inference_cfg: InferenceConfig,
    async_request_limiter: AsyncLimiter,
    stream: bool = False,
) -> AsyncStream[ChatCompletionChunk]:
    """Submit prompt to LLM and recieve stream of tokens as a result.
    inference_cfg: The configuration section from the config.json representing
    the relevant inference server for this request.
    """
    LOG.info("Analyzing the text")

    LOG.info("Submitting to /v1/chat/completions endpoint")

    async with async_request_limiter:
        response = await CLIENT.chat.completions.create(
            messages=messages,
            max_tokens=inference_cfg.max_tokens,
            logprobs=inference_cfg.log_probs,
            stream=stream,
            model=inference_cfg.model,
            temperature=inference_cfg.temperature,
        )

    return response


async def analyze_snippets(
    log_summary: List[Tuple[int, str]],
    async_request_limiter: AsyncLimiter,
    structured_output: dict | None = None,
) -> List[SnippetAnalysis | RatedSnippetAnalysis]:
    """Submit log file snippets to the LLM and gather results"""
    # Process snippets asynchronously
    awaitables = [
        call_llm(
            prompt_to_messages(
                PROMPT_CONFIG.snippet_prompt_template.format(s),
                PROMPT_CONFIG.snippet_system_prompt,
                SERVER_CONFIG.inference.system_role,
                SERVER_CONFIG.inference.user_role,
            ),
            async_request_limiter=async_request_limiter,
            inference_cfg=SERVER_CONFIG.snippet_inference,
            structured_output=structured_output,
            index=index
        )
        for index, s in enumerate(log_summary)
    ]
    gathered_responses = await asyncio.gather(*awaitables)
    analyzed_snippets = []

    for response in gathered_responses:
        if structured_output:
            try:
                snippet = RatedSnippetAnalysis.model_validate_json(response.text)
            except ValidationError as ex:
                LOG.error("Invalid data structure returned `%s`", response.text)
                raise ex
        else:
            snippet = SnippetAnalysis(text=response.text)
        analyzed_snippets.append(snippet)

    return analyzed_snippets


async def perform_analysis(
    log_text: str,
    async_request_limiter: AsyncLimiter,
    extractors: List[Extractor],
) -> Response:
    """Sumbit log file snippets in aggregate to LLM and retrieve results"""
    start_time = datetime.datetime.now()
    LOG.info("Starting analysis, Log length (in chars): %d", len(log_text))
    log_summary = mine_logs(log_text, extractors)
    log_summary = format_snippets(log_summary)

    final_prompt = construct_final_prompt(log_summary, PROMPT_CONFIG.prompt_template)

    messages = prompt_to_messages(
        final_prompt,
        PROMPT_CONFIG.default_system_prompt,
        SERVER_CONFIG.inference.system_role,
        SERVER_CONFIG.inference.user_role,
    )
    llm_start_time = datetime.datetime.now()
    response = await call_llm(
        messages,
        async_request_limiter=async_request_limiter,
        inference_cfg=SERVER_CONFIG.inference,
    )
    llm_timedelta = datetime.datetime.now() - llm_start_time
    LOG.info("Obtained LLM response, LLM call duration: %s", f"{llm_timedelta!s}")
    certainty = 0

    if response.logprobs is not None:
        try:
            certainty = compute_certainty(response.logprobs)
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n{response.logprobs}",
            ) from ex

    LOG.info(
        "Ending analysis, Analysis duration: %s",
        f"{(datetime.datetime.now() - start_time)!s}"
    )
    return Response(explanation=response, response_certainty=certainty)


async def perform_analysis_stream(
    log_text: str,
    async_request_limiter: AsyncLimiter,
    extractors: List[Extractor],
) -> AsyncStream:
    """Submit log file snippets in aggregate and return a stream of tokens"""
    start_time = datetime.datetime.now()
    LOG.info("Starting stream analysis, Log length (in chars): %d", len(log_text))

    log_summary = mine_logs(log_text, extractors)
    log_summary = format_snippets(log_summary)

    final_prompt = construct_final_prompt(log_summary, PROMPT_CONFIG.prompt_template)

    messages = prompt_to_messages(
        final_prompt,
        PROMPT_CONFIG.default_system_prompt,
        SERVER_CONFIG.inference.system_role,
        SERVER_CONFIG.inference.user_role,
    )

    llm_start_time = datetime.datetime.now()
    stream = call_llm_stream(
        messages,
        async_request_limiter=async_request_limiter,
        inference_cfg=SERVER_CONFIG.inference,
    )
    llm_duration = datetime.datetime.now() - llm_start_time
    LOG.info("Obtained LLM response, LLM stream call duration: %s", f"{llm_duration!s}")
    # we need to figure out a better response here, this is how it looks rn:
    # b'data: {"choices":[{"finish_reason":"stop","index":0,"delta":{}}],
    #   "created":1744818071,"id":"chatcmpl-c9geTxNcQO7M9wR...
    full_duration = datetime.datetime.now() - start_time
    LOG.info("Ending stream analysis, Analysis duration: %s", f"{full_duration!s}")
    return stream


# pylint: disable=too-many-locals
async def perform_staged_analysis(
    log_text: str,
    async_request_limiter: AsyncLimiter,
    extractors: List[Extractor],
) -> StagedResponse:
    """Submit the log file snippets to the LLM and retrieve their results"""
    start_time = datetime.datetime.now()
    LOG.info("Starting staged analysis, Log length (in chars): %d", len(log_text))
    log_summary = mine_logs(log_text, extractors)
    snippets_start_time = datetime.datetime.now()
    if SERVER_CONFIG.general.top_k_snippets:
        rated_snippets = await analyze_snippets(
            log_summary=log_summary,
            async_request_limiter=async_request_limiter,
            structured_output=RatedSnippetAnalysis.model_json_schema(),
        )

        # Extract original text and line number from `log_summary`
        processed_snippets = [
            AnalyzedSnippet(line_number=e[0][0], text=e[0][1], explanation=e[1])
            for e in zip(log_summary, rated_snippets)
        ]
        processed_snippets = filter_snippets(
            processed_snippets=processed_snippets,
            top_k=SERVER_CONFIG.general.top_k_snippets,
        )
        LOG.info(
            "Keeping %d of original %d snippets",
            len(processed_snippets),
            len(rated_snippets),
        )
    else:
        processed_snippets = await analyze_snippets(
            log_summary=log_summary, async_request_limiter=async_request_limiter
        )

        # Extract original text and line number from `log_summary`
        processed_snippets = [
            AnalyzedSnippet(line_number=e[0][0], text=e[0][1], explanation=e[1])
            for e in zip(log_summary, processed_snippets)
        ]
    snippet_analysis_duration = datetime.datetime.now() - snippets_start_time
    LOG.info("Snippet analysis duration: %s", f"{snippet_analysis_duration!s}")
    log_summary = format_analyzed_snippets(processed_snippets)
    final_prompt = construct_final_prompt(
        log_summary, PROMPT_CONFIG.prompt_template_staged
    )

    messages = prompt_to_messages(
        final_prompt,
        PROMPT_CONFIG.staged_system_prompt,
        SERVER_CONFIG.inference.system_role,
        SERVER_CONFIG.inference.user_role,
    )
    llm_start_time = datetime.datetime.now()
    LOG.info("Final LLM call on filtered snippets")
    final_analysis = await call_llm(
        messages,
        async_request_limiter=async_request_limiter,
        inference_cfg=SERVER_CONFIG.inference,
    )
    llm_duration = datetime.datetime.now() - llm_start_time
    LOG.info("Obtained LLM response, Final LLM call duration: %s", f"{llm_duration!s}")
    certainty = 0

    if final_analysis.logprobs:
        try:
            certainty = compute_certainty(final_analysis.logprobs)
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n"
                f"{final_analysis.logprobs}",
            ) from ex
    analysis_duration = datetime.datetime.now() - start_time
    LOG.info("Ending staged analysis, Analysis duration: %s", f"{analysis_duration!s}")
    return StagedResponse(
        explanation=final_analysis,
        snippets=processed_snippets,
        response_certainty=certainty,
    )



import inspect
from collections import defaultdict
import datetime
from typing import Optional, Union, Dict
from functools import wraps

import aiohttp
import numpy
from starlette.responses import StreamingResponse

from logdetective.remote_log import RemoteLog
from logdetective.server.config import LOG
from logdetective.server.compressors import LLMResponseCompressor, RemoteLogCompressor
from logdetective.server.models import (
    TimePeriod,
    MetricTimeSeries,
    StagedResponse,
    Response,
    Explanation,
)
from logdetective.server.database.models import EndpointType, AnalyzeRequestMetrics, Reactions
from logdetective.server.exceptions import LogDetectiveMetricsError


async def add_new_metrics(
    api_name: EndpointType,
    url: Optional[str] = None,
    http_session: Optional[aiohttp.ClientSession] = None,
    received_at: Optional[datetime.datetime] = None,
    compressed_log_content: Optional[bytes] = None,
) -> int:
    """Add a new database entry for a received request.

    This will store the time when this function is called,
    the endpoint from where the request was received,
    and the log (in a zip format) for which analysis is requested.
    """
    if not compressed_log_content:
        if not (url and http_session):
            raise LogDetectiveMetricsError(
                f"""Remote log can not be retrieved without URL and http session.
                URL: {url}, http session:{http_session}""")
        remote_log = RemoteLog(url, http_session)
        compressed_log_content = await RemoteLogCompressor(remote_log).zip_content()

    return await AnalyzeRequestMetrics.create(
        endpoint=EndpointType(api_name),
        compressed_log=compressed_log_content,
        request_received_at=received_at
        if received_at
        else datetime.datetime.now(datetime.timezone.utc),
    )


async def update_metrics(
    metrics_id: int,
    response: Union[Response, StagedResponse, StreamingResponse],
    sent_at: Optional[datetime.datetime] = None,
) -> None:
    """Update a database metric entry for a received request,
    filling data for the given response.

    This will add to the database entry the time when the response was sent,
    the length of the created response and the certainty for it.
    """
    try:
        compressed_response = LLMResponseCompressor(response).zip_response()
    except AttributeError as e:
        compressed_response = None
        LOG.warning(
            "Given response can not be serialized "
            "and saved in db (probably a StreamingResponse): %s.",
            e,
        )

    response_sent_at = (
        sent_at if sent_at else datetime.datetime.now(datetime.timezone.utc)
    )
    response_length = None
    if hasattr(response, "explanation") and isinstance(
        response.explanation, Explanation
    ):
        response_length = len(response.explanation.text)
    response_certainty = (
        response.response_certainty if hasattr(response, "response_certainty") else None
    )
    await AnalyzeRequestMetrics.update(
        id_=metrics_id,
        response_sent_at=response_sent_at,
        response_length=response_length,
        response_certainty=response_certainty,
        compressed_response=compressed_response,
    )


def track_request(name=None):
    """
    Decorator to track requests/responses metrics

    On entering the decorated function, it registers the time for the request
    and saves the passed log content.
    On exiting the decorated function, it registers the time for the response
    and saves the generated response.

    Use it to decorate server endpoints that generate a llm response
    as in the following example:

    >>> @app.post("/analyze", response_model=Response)
    >>> @track_request()
    >>> async def analyze_log(build_log)
    >>>     pass

    Warning: the decorators' order is important!
    The function returned by the *track_request* decorator is the
    server API function we want to be called by FastAPI.
    """

    def decorator(f):
        @wraps(f)
        async def async_decorated_function(*args, **kwargs):
            log_url = kwargs["build_log"].url
            metrics_id = await add_new_metrics(
                api_name=EndpointType(name if name else f.__name__),
                url=log_url, http_session=kwargs["http_session"]
            )
            response = await f(*args, **kwargs)
            await update_metrics(metrics_id, response)
            return response

        if inspect.iscoroutinefunction(f):
            return async_decorated_function
        raise NotImplementedError("An async coroutine is needed")

    return decorator


# TODO: Refactor aggregation to use database operations, instead of timestamp formatting  # pylint: disable=fixme
class TimeDefinition:
    """Define time format details, given a time period."""

    def __init__(self, time_period: TimePeriod):
        self.time_period = time_period
        self.days_diff = time_period.get_time_period().days
        if self.time_period.hours:
            self._time_format = "%Y-%m-%d %H"
            self._time_delta = datetime.timedelta(hours=1)
        elif self.time_period.days:
            self._time_format = "%Y-%m-%d"
            self._time_delta = datetime.timedelta(days=1)
        elif self.time_period.weeks:
            self._time_format = "%Y-%m-%d"
            self._time_delta = datetime.timedelta(weeks=1)

    @property
    def time_format(self):
        # pylint: disable=missing-function-docstring
        return self._time_format

    @property
    def time_delta(self):
        # pylint: disable=missing-function-docstring
        return self._time_delta


def create_time_series_arrays(
    values_dict: dict[datetime.datetime, int],
) -> tuple[list, list]:
    """Create time series arrays from a dictionary of values.

    This function generates two aligned lists:
    1. An array of timestamps from start_time to end_time
    2. A corresponding array of values for each timestamp

    Args:
        values_dict: Dictionary mapping timestamps to their respective values
    Returns:
        A tuple containing:
            - list: Array of timestamps
            - list: Array of corresponding values
    """

    timestamps = []
    values = []

    for timestamp, count in values_dict.items():
        timestamps.append(timestamp)
        values.append(count)

    return timestamps, numpy.nan_to_num(values).tolist()


async def requests_per_time(
    period_of_time: TimePeriod,
    endpoint: EndpointType = EndpointType.ANALYZE,
    end_time: Optional[datetime.datetime] = None,
) -> MetricTimeSeries:
    """
    Get request counts over a specified time period.

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        endpoint: One of the API endpoints
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A dictionary with timestamps and associated number of requests
    """
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    time_def = TimeDefinition(period_of_time)
    requests_counts = await AnalyzeRequestMetrics.get_requests_in_period(
        start_time, end_time, time_def.time_format, endpoint
    )
    timestamps, counts = create_time_series_arrays(requests_counts)

    return MetricTimeSeries(metric="requests", timestamps=timestamps, values=counts)


async def average_time_per_responses(
    period_of_time: TimePeriod,
    endpoint: EndpointType = EndpointType.ANALYZE,
    end_time: Optional[datetime.datetime] = None,
) -> MetricTimeSeries:
    """
    Get average response time and length over a specified time period.

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        endpoint: One of the API endpoints
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A dictionary of timestamps and average response times
    """
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    time_def = TimeDefinition(period_of_time)
    responses_average_time = (
        await AnalyzeRequestMetrics.get_responses_average_time_in_period(
            start_time, end_time, time_def.time_format, endpoint
        )
    )
    timestamps, average_time = create_time_series_arrays(
        responses_average_time,
    )

    return MetricTimeSeries(metric="avg_response_time", timestamps=timestamps, values=average_time)


async def _collect_emoji_data(
    start_time: datetime.datetime, time_def: TimeDefinition
) -> Dict[str, Dict[str, list]]:
    """Collect and organize emoji feedback data

    For each reaction type, a dictionary is created with time stamps
    as keys, and aggregate counts as values.
    """
    reactions = await Reactions.get_since(start_time)
    reaction_values: defaultdict[str, Dict] = defaultdict(lambda: defaultdict(int))

    for comment_timestamp, reaction in reactions:
        formatted_timestamp = comment_timestamp.strptime(
            comment_timestamp.strftime(time_def.time_format), time_def.time_format
        )

        reaction_values[reaction.reaction_type][formatted_timestamp] += reaction.count

    reaction_time_series = {
        reaction_type: {
            "timestamps": reaction_data.keys(),
            "values": reaction_data.values(),
        }
        for reaction_type, reaction_data in reaction_values.items()
    }

    return reaction_time_series


async def emojis_per_time(
    period_of_time: TimePeriod,
    end_time: Optional[datetime.datetime] = None,
) -> list[MetricTimeSeries]:
    """
    Retrieve data of emoji feedback over time.

    The time intervals are determined by the provided TimePeriod object, which defines
    the granularity.

    Args:
        period_of_time: A TimePeriod object that defines the time period and interval
                        for the analysis (e.g., hourly, daily, weekly)
        end_time: The end time for the analysis period. If None, defaults to the current
                  UTC time

    Returns:
        A list of `MetricTimeSeries` objects
    """
    time_def = TimeDefinition(period_of_time)
    end_time = end_time or datetime.datetime.now(datetime.timezone.utc)
    start_time = period_of_time.get_period_start_time(end_time)
    reactions_values_dict = await _collect_emoji_data(start_time, time_def)

    reaction_values: list[MetricTimeSeries] = []
    for reaction, time_series in reactions_values_dict.items():
        reaction_values.append(
            MetricTimeSeries(
                metric=f"emoji_{reaction}",
                timestamps=time_series["timestamps"],
                values=time_series["values"]))
    return reaction_values



"""This file contains various constants to be used as a fallback
in case other values are not specified. Prompt templates should be modified
in prompts.yaml instead.
"""

# pylint: disable=line-too-long
DEFAULT_ADVISOR = "fedora-copr/granite-3.2-8b-instruct-GGUF"

PROMPT_TEMPLATE = """
Given following log snippets, and nothing else, explain what failure, if any, occured during build of this package.

Analysis of the snippets must be in a format of [X] : [Y], where [X] is a log snippet, and [Y] is the explanation.
Snippets themselves must not be altered in any way whatsoever.

Snippets are delimited with '================'.

Finally, drawing on information from all snippets, provide complete explanation of the issue and recommend solution.

Explanation of the issue, and recommended solution, should take handful of sentences.

Snippets:

{}

Analysis:

"""

SNIPPET_PROMPT_TEMPLATE = """
Analyse following RPM build log snippet. Describe contents accurately, without speculation or suggestions for resolution.

Your analysis must be as concise as possible, while keeping relevant information intact.

Snippet:

{}

Analysis:

"""

PROMPT_TEMPLATE_STAGED = """
Given following log snippets, their explanation, and nothing else, explain what failure, if any, occured during build of this package.

Snippets are in a format of [X] : [Y], where [X] is a log snippet, and [Y] is the explanation.

Snippets are delimited with '================'.

Drawing on information from all snippets, provide complete explanation of the issue and recommend solution.

Explanation of the issue, and recommended solution, should take handful of sentences.

Snippets:

{}

Analysis:

"""

DEFAULT_SYSTEM_PROMPT = """
You are a highly capable large language model based expert system specialized in
packaging and delivery of software using RPM (RPM Package Manager). Your purpose is to diagnose
RPM build failures, identifying root causes and proposing solutions if possible.
You are truthful, concise, and helpful.

You never speculate about package being built or fabricate information.
If you do not know the answer, you acknowledge the fact and end your response.
Your responses must be as short as possible.
"""

SNIPPET_DELIMITER = "================"

DEFAULT_TEMPERATURE = 0.8

# Tuning for LLM-as-a-Service
LLM_DEFAULT_MAX_QUEUE_SIZE = 50
LLM_DEFAULT_REQUESTS_PER_MINUTE = 60

# Roles for chat API
SYSTEM_ROLE_DEFAULT = "developer"
USER_ROLE_DEFAULT = "user"



from __future__ import annotations
import enum
import datetime
from typing import Optional, List, Tuple, Self, TYPE_CHECKING

import backoff

from sqlalchemy import (
    Enum,
    BigInteger,
    DateTime,
    String,
    ForeignKey,
    UniqueConstraint,
    desc,
    select,
)
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.engine import Row
from sqlalchemy.exc import OperationalError
from logdetective.server.database.base import Base, transaction, DB_MAX_RETRIES


if TYPE_CHECKING:
    from .metrics import AnalyzeRequestMetrics


class Forge(str, enum.Enum):
    """List of forges managed by logdetective"""

    gitlab_com = "https://gitlab.com"  # pylint: disable=(invalid-name)
    gitlab_cee_redhat_com = "http://gitlab.cee.redhat.com/"  # pylint: disable=(invalid-name)


class GitlabMergeRequestJobs(Base):
    """Store details for the merge request jobs
    which triggered logdetective.
    """

    __tablename__ = "gitlab_merge_request_jobs"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    forge: Mapped[Forge] = mapped_column(
        Enum(Forge),
        nullable=False,
        index=True,
        comment="The forge name"
    )
    project_id: Mapped[int] = mapped_column(
        BigInteger,
        nullable=False,
        index=True,
        comment="The project gitlab id",
    )
    mr_iid: Mapped[int] = mapped_column(
        BigInteger,
        nullable=False,
        index=False,
        comment="The merge request gitlab iid",
    )
    job_id: Mapped[int] = mapped_column(
        BigInteger,
        nullable=False,
        index=True,
        comment="The job gitlab id",
    )

    __table_args__ = (
        UniqueConstraint("forge", "job_id", name="uix_forge_job"),
        UniqueConstraint(
            "forge", "project_id", "mr_iid", "job_id", name="uix_mr_project_job"
        ),
    )

    comment: Mapped[List["Comments"]] = relationship(
        "Comments", back_populates="merge_request_job", uselist=False
    )  # 1 comment for 1 job

    request_metrics: Mapped[List["AnalyzeRequestMetrics"]] = relationship(
        "AnalyzeRequestMetrics",
        back_populates="mr_job"
    )

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    async def create(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> int:
        """Create a new merge request job entry,
        returns its PostgreSQL id

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        async with transaction(commit=True) as session:
            mr = cls()
            mr.forge = forge
            mr.project_id = project_id
            mr.mr_iid = mr_iid
            mr.job_id = job_id
            session.add(mr)
            await session.flush()
            return mr.id

    @classmethod
    async def get_by_id(
        cls,
        id_: int,
    ) -> Optional["GitlabMergeRequestJobs"]:
        """Search for a given PostgreSQL id"""
        query = select(cls).where(cls.id == id_)
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            mr = query_result.scalars().first()
            return mr

    @classmethod
    async def get_by_details(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> Optional["GitlabMergeRequestJobs"]:
        """Search for a detailed merge request.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        query = select(cls).where(
            cls.forge == forge,
            cls.project_id == project_id,
            cls.mr_iid == mr_iid,
            cls.job_id == job_id,
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            mr = query_result.scalars().first()
            return mr

    @classmethod
    async def get_by_mr_iid(cls, forge: Forge, project_id: int, mr_iid) -> List[Self]:
        """Get all the mr jobs saved for the specified mr iid and project id."""
        query = select(cls).where(
            GitlabMergeRequestJobs.forge == forge,
            GitlabMergeRequestJobs.project_id == project_id,
            GitlabMergeRequestJobs.mr_iid == mr_iid,
        )

        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comments = query_result.scalars().all()

            return comments

    @classmethod
    async def get_or_create(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
    ) -> Self:
        """Search for a detailed merge request
        or create a new one if not found.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
        """
        mr = await GitlabMergeRequestJobs.get_by_details(
            forge, project_id, mr_iid, job_id
        )
        if mr is None:
            id_ = await GitlabMergeRequestJobs.create(forge, project_id, mr_iid, job_id)
            mr = await GitlabMergeRequestJobs.get_by_id(id_)
        return mr


class Comments(Base):
    """Store details for comments
    created by logdetective for a merge request job."""

    __tablename__ = "comments"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    merge_request_job_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("gitlab_merge_request_jobs.id"),
        nullable=False,
        unique=True,  # 1 comment for 1 job
        index=True,
        comment="The associated merge request job (db) id",
    )
    forge: Mapped[Forge] = mapped_column(
        Enum(Forge),
        nullable=False,
        index=True,
        comment="The forge name"
    )
    comment_id: Mapped[str] = mapped_column(
        String(50),  # e.g. 'd5a3ff139356ce33e37e73add446f16869741b50'
        nullable=False,
        index=True,
        comment="The comment gitlab id",
    )
    created_at: Mapped[datetime.datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        comment="Timestamp when the comment was created",
    )

    __table_args__ = (
        UniqueConstraint("forge", "comment_id", name="uix_forge_comment_id"),
    )

    merge_request_job: Mapped["GitlabMergeRequestJobs"] = relationship(
        "GitlabMergeRequestJobs",
        back_populates="comment"
    )
    reactions: Mapped[list["Reactions"]] = relationship("Reactions", back_populates="comment")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    async def create(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        metrics: Optional["AnalyzeRequestMetrics"] = None,  # noqa: F821
    ) -> int:
        """Create a new comment id entry,
        returns its PostgreSQL id.
        A Gitlab comment id is unique within a
        gitlab instance.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        async with transaction(commit=True) as session:
            mr_job = await GitlabMergeRequestJobs.get_or_create(
                forge, project_id, mr_iid, job_id
            )

            if metrics:
                metrics.mr_job = mr_job
                session.add(metrics)

            comment = cls()
            comment.forge = forge
            comment.comment_id = comment_id
            comment.created_at = datetime.datetime.now(datetime.timezone.utc)
            comment.merge_request_job_id = mr_job.id
            session.add(comment)
            await session.flush()
            return comment.id

    @classmethod
    async def get_by_id(
        cls,
        id_: int,
    ) -> Optional["Comments"]:
        """Search for a given PostgreSQL id"""
        query = select(cls).where(cls.id == id_)
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comment = query_result.scalars().first()
            return comment

    @classmethod
    async def get_by_gitlab_id(
        cls,
        forge: Forge,
        comment_id: str,
    ) -> Optional[Self]:
        """Search for a detailed comment
        by its unique forge comment id.

        Args:
          forge: forge name
          comment_id: forge comment id
        """
        query = (
            select(cls)
            .join(
                GitlabMergeRequestJobs,
                cls.merge_request_job_id == GitlabMergeRequestJobs.id,
            )
            .filter(GitlabMergeRequestJobs.forge == forge, cls.comment_id == comment_id)
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comment = query_result.scalars().first()
            return comment

    @classmethod
    async def get_latest_comment(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
    ) -> Optional["Comments"]:
        """Search for the latest comment in the merge request.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
        """
        query = (
            select(cls)
            .join(
                GitlabMergeRequestJobs,
                cls.merge_request_job_id == GitlabMergeRequestJobs.id,
            )
            .filter(
                GitlabMergeRequestJobs.forge == forge,
                GitlabMergeRequestJobs.project_id == project_id,
                GitlabMergeRequestJobs.mr_iid == mr_iid,
            )
            .order_by(desc(cls.created_at))
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comment = query_result.scalars().first()
            return comment

    @classmethod
    async def get_mr_comments(
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
    ) -> List[Self]:
        """Search for all merge request comments.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
        """
        query = (
            select(cls)
            .join(
                GitlabMergeRequestJobs,
                cls.merge_request_job_id == GitlabMergeRequestJobs.id,
            )
            .filter(
                GitlabMergeRequestJobs.forge == forge,
                GitlabMergeRequestJobs.project_id == project_id,
                GitlabMergeRequestJobs.mr_iid == mr_iid,
            )
            .order_by(desc(cls.created_at))
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comments = query_result.scalars().all()
            return comments

    @classmethod
    async def get_or_create(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
    ) -> Self:
        """Search for a detailed comment
        or create a new one if not found.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        comment = await Comments.get_by_gitlab_id(forge, comment_id)
        if comment is None:
            id_ = await Comments.create(forge, project_id, mr_iid, job_id, comment_id)
            comment = await Comments.get_by_id(id_)
        return comment

    @classmethod
    async def get_since(cls, time: datetime.datetime) -> List[Self]:
        """Get all the comments created after the given time."""
        query = select(cls).filter(Comments.created_at > time)
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comments = query_result.scalars().all()

            return comments

    @classmethod
    async def get_by_mr_job(
        cls, merge_request_job: GitlabMergeRequestJobs
    ) -> Optional["Comments"]:
        """Get the comment added for the specified merge request's job."""
        query = select(cls).filter(Comments.merge_request_job == merge_request_job)
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            comments = query_result.scalars().first()
            return comments


class Reactions(Base):
    """Store and count reactions received for
    logdetective comments"""

    __tablename__ = "reactions"

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    comment_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("comments.id"),
        nullable=False,
        index=True,
        comment="The associated comment (db) id",
    )
    reaction_type: Mapped[str] = mapped_column(
        String(127),  # e.g. 'thumbs-up'
        nullable=False,
        comment="The type of reaction",
    )
    count: Mapped[int] = mapped_column(
        BigInteger,
        nullable=False,
        comment="The number of reactions, of this type, given in the comment",
    )

    __table_args__ = (
        UniqueConstraint("comment_id", "reaction_type", name="uix_comment_reaction"),
    )

    comment: Mapped["Comments"] = relationship("Comments", back_populates="reactions")

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    async def create_or_update(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        reaction_type: str,
        count: int,
    ) -> int:
        """Create or update the given reaction, and its associated count,
        for a specified comment.

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
          reaction_type: a str, ex. thumb_up
          count: number of reactions, of this type, given in the comment
        """
        comment = await Comments.get_or_create(
            forge, project_id, mr_iid, job_id, comment_id
        )
        reaction = await cls.get_reaction_by_type(
            forge, project_id, mr_iid, job_id, comment_id, reaction_type
        )
        async with transaction(commit=True) as session:
            if reaction:
                reaction.count = count  # just update
            else:
                reaction = cls()
                reaction.comment_id = comment.id
                reaction.reaction_type = reaction_type
                reaction.count = count
            session.add(reaction)
            await session.flush()
            return reaction.id

    @classmethod
    async def get_all_reactions(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
    ) -> List[Self]:
        """Get all reactions for a comment

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
        """
        query = (
            select(cls)
            .join(Comments, cls.comment_id == Comments.id)
            .join(
                GitlabMergeRequestJobs,
                Comments.merge_request_job_id == GitlabMergeRequestJobs.id,
            )
            .filter(
                Comments.comment_id == comment_id,
                GitlabMergeRequestJobs.forge == forge,
                GitlabMergeRequestJobs.project_id == project_id,
                GitlabMergeRequestJobs.mr_iid == mr_iid,
                GitlabMergeRequestJobs.job_id == job_id,
            )
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            reactions = query_result.scalars().all()
            return reactions

    @classmethod
    async def get_reaction_by_type(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        reaction_type: str,
    ) -> Self | None:
        """Get reaction, of a given type,
        for a comment

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
          reaction_type: str like "thumb-up"
        """
        query = (
            select(cls)
            .join(Comments, cls.comment_id == Comments.id)
            .join(
                GitlabMergeRequestJobs,
                Comments.merge_request_job_id == GitlabMergeRequestJobs.id,
            )
            .filter(
                Comments.comment_id == comment_id,
                GitlabMergeRequestJobs.forge == forge,
                GitlabMergeRequestJobs.project_id == project_id,
                GitlabMergeRequestJobs.mr_iid == mr_iid,
                GitlabMergeRequestJobs.job_id == job_id,
                Reactions.reaction_type == reaction_type,
            )
        )
        async with transaction(commit=False) as session:
            query_result = await session.execute(query)
            reaction = query_result.scalars().first()
            return reaction

    @classmethod
    @backoff.on_exception(backoff.expo, OperationalError, max_tries=DB_MAX_RETRIES)
    async def delete(  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
        cls,
        forge: Forge,
        project_id: int,
        mr_iid: int,
        job_id: int,
        comment_id: str,
        reaction_types: List[str],
    ) -> None:
        """Remove rows with given reaction types

        Args:
          forge: forge name
          project_id: forge project id
          mr_iid: merge request forge iid
          job_id: forge job id
          comment_id: forge comment id
          reaction_type: a str iterable, ex. ['thumbsup', 'thumbsdown']
        """
        for reaction_type in reaction_types:
            reaction = await cls.get_reaction_by_type(
                forge, project_id, mr_iid, job_id, comment_id, reaction_type
            )
            async with transaction(commit=True) as session:
                await session.delete(reaction)
                await session.flush()

    @classmethod
    async def get_since(
        cls, time: datetime.datetime
    ) -> List[Row[Tuple[datetime.datetime, Self]]]:
        """Get all the reactions on comments created after the given time
        and the comment creation time."""
        query = (
            select(Comments.created_at, cls)
            .join(Comments, cls.comment_id == Comments.id)
            .filter(Comments.created_at > time)
        )

        async with transaction(commit=False) as session:
            query_results = await session.execute(query)
            reactions = query_results.all()

            return reactions



import os
import asyncio
import datetime
from enum import Enum
from collections import defaultdict
from contextlib import asynccontextmanager
from typing import Annotated

from aiolimiter import AsyncLimiter
from koji import ClientSession
from gitlab import Gitlab
from fastapi import (
    FastAPI,
    HTTPException,
    BackgroundTasks,
    Depends,
    Header,
    Path,
    Request,
)
from fastapi.responses import StreamingResponse
from fastapi.responses import Response as BasicResponse
import aiohttp
import sentry_sdk

from logdetective.extractors import DrainExtractor, CSGrepExtractor, Extractor
from logdetective.utils import sanitize_log
from logdetective.server.exceptions import KojiInvalidTaskID

from logdetective.server.database.models.koji import KojiTaskAnalysis
from logdetective.server.database.models.exceptions import (
    KojiTaskAnalysisTimeoutError,
    KojiTaskNotAnalyzedError,
    KojiTaskNotFoundError,
)

import logdetective.server.database.base

from logdetective.server.config import SERVER_CONFIG, LOG
from logdetective.server.koji import (
    get_failed_log_from_task as get_failed_log_from_koji_task,
)
from logdetective.remote_log import RemoteLog
from logdetective.server.llm import (
    perform_staged_analysis,
    perform_analysis,
    perform_analysis_stream,
)
from logdetective.server.gitlab import process_gitlab_job_event
from logdetective.server.metric import (
    track_request,
    add_new_metrics,
    update_metrics,
    requests_per_time,
    average_time_per_responses,
    emojis_per_time
)
from logdetective.server.models import (
    BuildLog,
    Config,
    EmojiHook,
    JobHook,
    KojiInstanceConfig,
    KojiStagedResponse,
    Response,
    StagedResponse,
    TimePeriod,
    ExtractorConfig,
    MetricResponse,
)
from logdetective.server.database.models import (
    EndpointType,
    Forge,
)
from logdetective.server.emoji import (
    collect_emojis,
    collect_emojis_for_mr,
)
from logdetective.server.compressors import RemoteLogCompressor
from logdetective.server.utils import get_version


LOG_SOURCE_REQUEST_TIMEOUT = os.environ.get("LOG_SOURCE_REQUEST_TIMEOUT", 60)
API_TOKEN = os.environ.get("LOGDETECTIVE_TOKEN", None)


if sentry_dsn := SERVER_CONFIG.general.sentry_dsn:
    sentry_sdk.init(dsn=str(sentry_dsn), traces_sample_rate=1.0)


def initialize_extractors(extractor_config: ExtractorConfig) -> list[Extractor]:
    """Set up extractors based on provided ExtractorConfig."""
    extractors: list[Extractor] = [
        DrainExtractor(
            verbose=extractor_config.verbose,
            max_snippet_len=extractor_config.max_snippet_len,
            max_clusters=extractor_config.max_clusters,
        )
    ]

    if extractor_config.csgrep:
        extractors.append(
            CSGrepExtractor(
                verbose=extractor_config.verbose,
                max_snippet_len=extractor_config.max_snippet_len,
            )
        )

    return extractors


class ConnectionManager:
    """
    Manager for all connections and sesssions.
    """

    koji_connections: dict[str, ClientSession] = {}
    gitlab_connections: dict[str, Gitlab] = {}
    gitlab_http_sessions: dict[str, aiohttp.ClientSession] = {}

    async def initialize(self, service_config: Config):
        """Initialize all managed objects"""

        for connection, config in service_config.gitlab.instances.items():
            self.gitlab_connections[connection] = Gitlab(
                url=config.url,
                private_token=config.api_token,
                timeout=config.timeout,
            )
            self.gitlab_http_sessions[connection] = aiohttp.ClientSession(
                base_url=config.url,
                headers={"Authorization": f"Bearer {config.api_token}"},
                timeout=aiohttp.ClientTimeout(
                    total=config.timeout,
                    connect=3.07,
                ),
            )
        for connection, config in service_config.koji.instances.items():
            self.koji_connections[connection] = ClientSession(baseurl=config.xmlrpc_url)

    async def close(self):
        """Close all managed http sessions"""
        for session in self.gitlab_http_sessions.values():

            await session.close()


class KojiCallbackManager:
    """Manages callbacks used by Koji, with callbacks referenced by task id.

    Multiple callbacks can be assigned to a single task."""

    _callbacks: defaultdict[int, set[str]]

    def __init__(self) -> None:
        self._callbacks = defaultdict(set)

    def register_callback(self, task_id: int, callback: str):
        """Register a callback for a task"""
        self._callbacks[task_id].add(callback)

    def clear_callbacks(self, task_id: int):
        """Unregister a callback for a task"""
        try:
            del self._callbacks[task_id]
        except KeyError:
            pass

    def get_callbacks(self, task_id: int) -> set[str]:
        """Get the callbacks for a task"""
        return self._callbacks[task_id]


@asynccontextmanager
async def lifespan(fapp: FastAPI):
    """
    Establish one HTTP session
    """
    fapp.http = aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(
            total=int(LOG_SOURCE_REQUEST_TIMEOUT), connect=3.07
        )
    )

    # General limiter for async requests
    fapp.state.llm_request_limiter = AsyncLimiter(
        max_rate=SERVER_CONFIG.inference.requests_per_minute
    )

    # Manager for connections and sessions
    fapp.state.connection_manager = ConnectionManager()

    await fapp.state.connection_manager.initialize(service_config=SERVER_CONFIG)

    # Set up extractors
    fapp.state.extractors = initialize_extractors(SERVER_CONFIG.extractor)

    # Koji callbacks
    fapp.state.koji_callback_manager = KojiCallbackManager()

    # Ensure that the database is initialized.
    await logdetective.server.database.base.init()

    # Start the background task scheduler for collecting emojis
    asyncio.create_task(schedule_collect_emojis_task(fapp.state.connection_manager))

    yield

    await fapp.state.connection_manager.close()
    await fapp.http.close()


async def get_http_session(request: Request) -> aiohttp.ClientSession:
    """
    Return the single aiohttp ClientSession for this app
    """
    return request.app.http


def requires_token_when_set(authorization: Annotated[str | None, Header()] = None):
    """
    FastAPI Depend function that expects a header named Authorization

    If LOGDETECTIVE_TOKEN env var is set, validate the client-supplied token
    otherwise ignore it
    """
    if not API_TOKEN:
        LOG.info("LOGDETECTIVE_TOKEN env var not set, authorization disabled")
        # no token required, means local dev environment
        return
    if authorization:
        try:
            token = authorization.split(" ", 1)[1]
        except (ValueError, IndexError) as ex:
            LOG.warning(
                "Authorization header has invalid structure '%s', it should be 'Bearer TOKEN'",
                authorization,
            )
            # eat the exception and raise 401 below
            raise HTTPException(
                status_code=401,
                detail=f"Invalid authorization, HEADER '{authorization}' not valid.",
            ) from ex
        if token == API_TOKEN:
            return
        LOG.info("Provided token '%s' does not match expected value.", token)
        raise HTTPException(status_code=401, detail=f"Token '{token}' not valid.")
    LOG.error("No authorization header provided but LOGDETECTIVE_TOKEN env var is set")
    raise HTTPException(status_code=401, detail="No token provided.")


app = FastAPI(
    title="Log Detective",
    contact={
        "name": "Log Detective developers",
        "url": "https://github.com/fedora-copr/logdetective",
        "email": "copr-devel@lists.fedorahosted.org",
    },
    license_info={
        "name": "Apache-2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    },
    version=get_version(),
    dependencies=[Depends(requires_token_when_set)],
    lifespan=lifespan,
)


@app.post("/analyze", response_model=Response)
@track_request()
async def analyze_log(
    build_log: BuildLog,
    request: Request,
    http_session: aiohttp.ClientSession = Depends(get_http_session),
):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()
    log_text = sanitize_log(log_text)

    return await perform_analysis(
        log_text,
        async_request_limiter=request.app.state.llm_request_limiter,
        extractors=request.app.state.extractors
    )


@app.post("/analyze/staged", response_model=StagedResponse)
@track_request()
async def analyze_log_staged(
    build_log: BuildLog,
    request: Request,
    http_session: aiohttp.ClientSession = Depends(get_http_session),
):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()
    log_text = sanitize_log(log_text)

    return await perform_staged_analysis(
        log_text,
        async_request_limiter=request.app.state.llm_request_limiter,
        extractors=request.app.state.extractors,
    )


@app.get(
    "/analyze/rpmbuild/koji/{koji_instance}/{task_id}",
    response_model=KojiStagedResponse,
)
async def get_koji_task_analysis(
    koji_instance: Annotated[str, Path(title="The Koji instance to use")],
    task_id: Annotated[int, Path(title="The task ID to analyze")],
    x_koji_token: Annotated[str, Header()] = "",
):
    """Provide endpoint for retrieving log file analysis of a Koji task"""

    try:
        koji_instance_config = SERVER_CONFIG.koji.instances[koji_instance]
    except KeyError:
        # This Koji instance is not configured, so we will return a 404.
        return BasicResponse(status_code=404, content="Unknown Koji instance.")

    # This should always be available in a production environment.
    # In a testing environment, the tokens list may be empty, in which case
    # it will just proceed.
    if koji_instance_config.tokens and x_koji_token not in koji_instance_config.tokens:
        # (Unauthorized) error.
        return BasicResponse(x_koji_token, status_code=401)

    # Check if we have a response for this task
    try:
        return KojiTaskAnalysis.get_response_by_task_id(task_id)

    except (KojiInvalidTaskID, KojiTaskNotFoundError):
        # This task ID is malformed, out of range, or not found, so we will
        # return a 404.
        return BasicResponse(status_code=404)

    except KojiTaskAnalysisTimeoutError:
        # Task analysis has timed out, so we assume that the request was lost
        # and that we need to start another analysis.
        # There isn't a fully-appropriate error code for this, so we'll use
        # 503 (Service Unavailable) as our best option.
        return BasicResponse(
            status_code=503, content="Task analysis timed out, please retry."
        )

    except KojiTaskNotAnalyzedError:
        # Its still running, so we need to return a 202
        # (Accepted) code to let the client know to keep waiting.
        return BasicResponse(
            status_code=202, content=f"Analysis still in progress for task {task_id}"
        )


@app.post(
    "/analyze/rpmbuild/koji/{koji_instance}/{task_id}",
    response_model=KojiStagedResponse,
)
async def analyze_rpmbuild_koji(
    koji_instance: Annotated[str, Path(title="The Koji instance to use")],
    task_id: Annotated[int, Path(title="The task ID to analyze")],
    request: Request,
    x_koji_token: Annotated[str, Header()] = "",
    x_koji_callback: Annotated[str, Header()] = "",
    background_tasks: BackgroundTasks = BackgroundTasks(),
):  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
    """Provide endpoint for retrieving log file analysis of a Koji task"""

    try:
        koji_instance_config = SERVER_CONFIG.koji.instances[koji_instance]
    except KeyError:
        # This Koji instance is not configured, so we will return a 404.
        return BasicResponse(status_code=404, content="Unknown Koji instance.")

    # This should always be available in a production environment.
    # In a testing environment, the tokens list may be empty, in which case
    # it will just proceed.
    if koji_instance_config.tokens and x_koji_token not in koji_instance_config.tokens:
        # (Unauthorized) error.
        return BasicResponse(x_koji_token, status_code=401)

    # Check if we already have a response for this task
    try:
        response = KojiTaskAnalysis.get_response_by_task_id(task_id)

    except KojiInvalidTaskID:
        # This task ID is malformed or out of range, so we will return a 400.
        response = BasicResponse(status_code=404, content="Invalid or unknown task ID.")

    except (KojiTaskNotFoundError, KojiTaskAnalysisTimeoutError):
        # Task not yet analyzed or it timed out, so we need to start the
        # analysis in the background and return a 202 (Accepted) error.

        koji_connection = request.app.state.connection_manager.koji_connections[
            koji_instance
        ]
        background_tasks.add_task(
            analyze_koji_task,
            task_id,
            koji_instance_config,
            koji_connection,
            request.app.state.llm_request_limiter,
            request.app.state.extractors,
            request.app.state.koji_callback_manager
        )

        # If a callback URL is provided, we need to add it to the callbacks
        # table so that we can notify it when the analysis is complete.
        if x_koji_callback:
            request.app.state.koji_callback_manager.register_callback(task_id, x_koji_callback)

        response = BasicResponse(
            status_code=202, content=f"Beginning analysis of task {task_id}"
        )

    except KojiTaskNotAnalyzedError:
        # Its still running, so we need to return a 202
        # (Accepted) error.
        response = BasicResponse(
            status_code=202, content=f"Analysis still in progress for task {task_id}"
        )

    return response


async def analyze_koji_task(
    task_id: int,
    koji_instance_config: KojiInstanceConfig,
    koji_connection: ClientSession,
    async_request_limiter: AsyncLimiter,
    extractors: list[Extractor],
    koji_callback_manager: KojiCallbackManager,
):  # pylint: disable=too-many-arguments disable=too-many-positional-arguments
    """Analyze a koji task and return the response"""

    # Get the log text from the koji task
    log_file_name, log_text = await get_failed_log_from_koji_task(
        koji_connection, task_id, max_size=SERVER_CONFIG.koji.max_artifact_size
    )
    log_text = sanitize_log(log_text)

    # We need to handle the metric tracking manually here, because we need
    # to retrieve the metric ID to associate it with the koji task analysis.

    metrics_id = await add_new_metrics(
        EndpointType.ANALYZE_KOJI_TASK,
        log_text,
        received_at=datetime.datetime.now(datetime.timezone.utc),
        compressed_log_content=RemoteLogCompressor.zip_text(log_text),
    )
    # We need to associate the metric ID with the koji task analysis.
    # This will create the new row without a response, which we will use as
    # an indicator that the analysis is in progress.
    await KojiTaskAnalysis.create_or_restart(
        koji_instance=koji_instance_config.xmlrpc_url,
        task_id=task_id,
        log_file_name=log_file_name,
    )
    response = await perform_staged_analysis(
        log_text,
        async_request_limiter=async_request_limiter,
        extractors=extractors
    )

    # Now that we have the response, we can update the metrics and mark the
    # koji task analysis as completed.
    await update_metrics(metrics_id, response)
    await KojiTaskAnalysis.add_response(task_id, metrics_id)

    # Notify any callbacks that the analysis is complete.
    for callback in koji_callback_manager.get_callbacks(task_id):
        LOG.info("Notifying callback %s of task %d completion", callback, task_id)
        asyncio.create_task(send_koji_callback(callback, task_id))

    # Now that it's sent, we can clear the callbacks for this task.
    koji_callback_manager.clear_callbacks(task_id)

    return response


async def send_koji_callback(callback: str, task_id: int):
    """Send a callback to the specified URL with the task ID and log file name."""
    async with aiohttp.ClientSession() as session:
        async with session.post(callback, json={"task_id": task_id}):
            pass


@app.get("/queue/print")
async def queue_print(msg: str, request: Request):
    """Debug endpoint to test the LLM request queue"""
    LOG.info("Will print %s", msg)

    result = await async_log(msg, request)

    LOG.info("Printed %s and returned it", result)

    return result


async def async_log(msg: str, request: Request):
    """Debug function to test the LLM request queue"""
    async with request.app.state.llm_request_limiter:
        LOG.critical(msg)
    return msg


@app.get("/version", response_class=BasicResponse)
async def get_version_wrapper():
    """Get the version of logdetective"""
    return BasicResponse(content=get_version())


@app.post("/analyze/stream", response_class=StreamingResponse)
@track_request()
async def analyze_log_stream(
    build_log: BuildLog,
    request: Request,
    http_session: aiohttp.ClientSession = Depends(get_http_session),
):
    """Stream response endpoint for Logdetective.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    remote_log = RemoteLog(build_log.url, http_session)
    log_text = await remote_log.process_url()
    log_text = sanitize_log(log_text)
    try:
        stream = perform_analysis_stream(
            log_text,
            async_request_limiter=request.app.state.llm_request_limiter,
            extractors=request.app.state.extractors,
        )
    except aiohttp.ClientResponseError as ex:
        raise HTTPException(
            status_code=400,
            detail="HTTP Error while getting response from inference server "
            f"[{ex.status}] {ex.message}",
        ) from ex

    return StreamingResponse(stream)


def is_valid_webhook_secret(forge, x_gitlab_token):
    """Check whether the provided x_gitlab_token matches the webhook secret
    specified in the configuration"""

    gitlab_cfg = SERVER_CONFIG.gitlab.instances[forge.value]

    if not gitlab_cfg.webhook_secrets:
        # No secrets specified, so don't bother validating.
        # This is mostly to be used for development.
        return True

    if x_gitlab_token in gitlab_cfg.webhook_secrets:
        return True

    return False


@app.post("/webhook/gitlab/job_events")
async def receive_gitlab_job_event_webhook(
    job_hook: JobHook,
    background_tasks: BackgroundTasks,
    request: Request,
    x_gitlab_instance: Annotated[str | None, Header()],
    x_gitlab_token: Annotated[str | None, Header()] = None,
):
    """Webhook endpoint for receiving job_events notifications from GitLab
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    lists the full specification for the messages sent for job events."""

    try:
        forge = Forge(x_gitlab_instance)
    except ValueError:
        LOG.critical("%s is not a recognized forge. Ignoring.", x_gitlab_instance)
        return BasicResponse(status_code=400)

    if not is_valid_webhook_secret(forge, x_gitlab_token):
        # This request could not be validated, so return a 401
        # (Unauthorized) error.
        return BasicResponse(status_code=401)

    # Handle the message in the background so we can return 204 immediately
    gitlab_cfg = SERVER_CONFIG.gitlab.instances[forge.value]
    gitlab_connection = request.app.state.connection_manager.gitlab_connections[
        forge.value
    ]
    gitlab_http_session = request.app.state.connection_manager.gitlab_http_sessions[
        forge.value
    ]
    background_tasks.add_task(
        process_gitlab_job_event,
        gitlab_cfg,
        gitlab_connection,
        gitlab_http_session,
        forge,
        job_hook,
        request.app.state.llm_request_limiter,
        request.app.state.extractors,
        SERVER_CONFIG.general.report_certainty,
    )

    # No return value or body is required for a webhook.
    # 204: No Content
    return BasicResponse(status_code=204)


# A lookup table for whether we are currently processing a given merge request
# The key is the tuple (Forge, ProjectID, MRID) and the value is a boolean
# indicating whether we need to re-trigger the lookup immediately after
# completion due to another request coming in during processing.
# For example: {("https://gitlab.example.com", 23, 2): False}
emoji_lookup = {}


@app.post("/webhook/gitlab/emoji_events")
async def receive_gitlab_emoji_event_webhook(
    x_gitlab_instance: Annotated[str | None, Header()],
    x_gitlab_token: Annotated[str | None, Header()],
    emoji_hook: EmojiHook,
    background_tasks: BackgroundTasks,
    request: Request,
):
    """Webhook endpoint for receiving emoji event notifications from Gitlab
    https://docs.gitlab.com/user/project/integrations/webhook_events/#emoji-events
    lists the full specification for the messages sent for emoji events"""

    try:
        forge = Forge(x_gitlab_instance)
    except ValueError:
        LOG.critical("%s is not a recognized forge. Ignoring.", x_gitlab_instance)
        return BasicResponse(status_code=400)

    if not is_valid_webhook_secret(forge, x_gitlab_token):
        # This request could not be validated, so return a 401
        # (Unauthorized) error.
        return BasicResponse(status_code=401)

    if not emoji_hook.merge_request:
        # This is not a merge request event. It is probably an emoji applied
        # to some other "awardable" entity. Just ignore it and return.
        LOG.debug("Emoji event is not related to a merge request. Ignoring.")
        return BasicResponse(status_code=204)

    # We will re-process all the emojis on this merge request, to ensure that
    # we haven't missed any messages, since webhooks do not provide delivery
    # guarantees.

    # Check whether this request is already in progress.
    # We are single-threaded, so we can guarantee that the table won't change
    # between here and when we schedule the lookup.
    key = (
        forge,
        emoji_hook.merge_request.target_project_id,
        emoji_hook.merge_request.iid,
    )
    if key in emoji_lookup:
        # It's already in progress, so we do not want to start another pass
        # concurrently. We'll set the value to True to indicate that we should
        # re-enqueue this lookup after the currently-running one concludes. It
        # is always safe to set this to True, even if it's already True. If
        # multiple requests come in during processing, we only need to re-run
        # it a single time, since it will pick up all the ongoing changes. The
        # worst-case situation is the one where we receive new requests just
        # after processing starts, which will cause the cycle to repeat again.
        # This should be very infrequent, as emoji events are computationally
        # rare and very quick to process.
        emoji_lookup[key] = True
        LOG.info("MR Emojis already being processed for %s. Rescheduling.", key)
        return BasicResponse(status_code=204)

    # Inform the lookup table that we are processing this emoji
    emoji_lookup[key] = False

    gitlab_connection = request.app.state.connection_manager.gitlab_connections[
        x_gitlab_instance
    ]
    # Create a background task to process the emojis on this Merge Request.
    background_tasks.add_task(
        schedule_emoji_collection_for_mr,
        forge,
        gitlab_connection,
        emoji_hook.merge_request.target_project_id,
        emoji_hook.merge_request.iid,
        background_tasks,
    )

    # No return value or body is required for a webhook.
    # 204: No Content
    return BasicResponse(status_code=204)


async def schedule_emoji_collection_for_mr(
    forge: Forge,
    gitlab_connection: Gitlab,
    project_id: int,
    mr_iid: int,
    background_tasks: BackgroundTasks,
):
    """Background task to update the database on emoji reactions"""

    key = (forge, project_id, mr_iid)

    # FIXME: Look up the connection from the Forge  # pylint: disable=fixme
    # gitlab_conn = SERVER_CONFIG.gitlab.instances[forge.value].get_connection()

    LOG.debug("Looking up emojis for %s, %d, %d", forge, project_id, mr_iid)
    await collect_emojis_for_mr(project_id, mr_iid, gitlab_connection)

    # Check whether we've been asked to re-schedule this lookup because
    # another request came in while it was processing.
    if emoji_lookup[key]:
        # The value is Truthy, which tells us to re-schedule
        # Reset the boolean value to indicate that we're underway again.
        emoji_lookup[key] = False
        background_tasks.add_task(
            schedule_emoji_collection_for_mr,
            forge,
            gitlab_connection,
            project_id,
            mr_iid,
            background_tasks,
        )
        return

    # We're all done, so clear this entry out of the lookup
    del emoji_lookup[key]


class MetricRoute(str, Enum):
    """Routes for metrics"""

    ANALYZE = "analyze"
    ANALYZE_STAGED = "analyze-staged"
    ANALYZE_GITLAB_JOB = "analyze-gitlab"


class MetricType(str, Enum):
    """Type of metric retrieved"""

    REQUESTS = "requests"
    RESPONSES = "responses"
    EMOJIS = "emojis"
    ALL = "all"


ROUTE_TO_ENDPOINT_TYPES = {
    MetricRoute.ANALYZE: EndpointType.ANALYZE,
    MetricRoute.ANALYZE_STAGED: EndpointType.ANALYZE_STAGED,
    MetricRoute.ANALYZE_GITLAB_JOB: EndpointType.ANALYZE_GITLAB_JOB,
}


@app.get("/metrics/{route}/", response_model=MetricResponse)
@app.get("/metrics/{route}/{metric_type}", response_model=MetricResponse)
async def get_metrics(
    route: MetricRoute,
    metric_type: MetricType = MetricType.ALL,
    period_since_now: TimePeriod = Depends(TimePeriod),
):
    """Get an handler returning statistics for the specified endpoint and metric_type."""
    endpoint_type = ROUTE_TO_ENDPOINT_TYPES[route]

    async def handler() -> MetricResponse:
        """Return statistics for the specified endpoint and metric type."""
        statistics = []
        if metric_type == MetricType.ALL:
            statistics.append(await requests_per_time(
                period_since_now, endpoint_type
            ))
            statistics.append(await average_time_per_responses(
                period_since_now, endpoint_type
            ))
            statistics.extend(await emojis_per_time(period_since_now))
            return MetricResponse(time_series=statistics)
        if metric_type == MetricType.REQUESTS:
            statistics.append(await requests_per_time(period_since_now, endpoint_type))
        elif metric_type == MetricType.RESPONSES:
            statistics.append(await average_time_per_responses(
                period_since_now, endpoint_type
            ))
        elif metric_type == MetricType.EMOJIS:
            statistics = await emojis_per_time(period_since_now)
        return MetricResponse(time_series=statistics)

    descriptions = {
        MetricType.REQUESTS: (
            "Get statistics for the requests received in the given period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
        MetricType.RESPONSES: (
            "Get statistics for responses given in the specified period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
        MetricType.EMOJIS: (
            "Get statistics for emoji feedback in the specified period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
        MetricType.ALL: (
            "Get statistics for requests and responses in the given period of time "
            f"for the /{endpoint_type.value} API endpoint."
        ),
    }
    handler.__doc__ = descriptions[metric_type]

    return await handler()


async def collect_emoji_task(connection_manager: ConnectionManager):
    """Collect emoji feedback.
    Query only comments created in the last year.
    """

    for instance_url, instance in SERVER_CONFIG.gitlab.instances.items():
        LOG.info(
            "Collect emoji feedback for %s started at %s",
            instance.url,
            datetime.datetime.now(datetime.timezone.utc),
        )
        await collect_emojis(
            connection_manager.gitlab_connections[instance_url], TimePeriod(weeks=54)
        )
        LOG.info(
            "Collect emoji feedback finished at %s",
            datetime.datetime.now(datetime.timezone.utc),
        )


async def schedule_collect_emojis_task(connection_manager: ConnectionManager):
    """Schedule the collect_emojis_task to run on a configured interval"""
    while True:
        seconds_until_run = SERVER_CONFIG.general.collect_emojis_interval
        LOG.info("Collect emojis in %d seconds", seconds_until_run)
        await asyncio.sleep(seconds_until_run)

        try:
            await collect_emoji_task(connection_manager=connection_manager)
        except Exception as e:  # pylint: disable=broad-exception-caught
            LOG.exception("Error in collect_emoji_task: %s", e)



import datetime
from typing import List, Dict, Optional, Any
from pydantic import (
    BaseModel,
    Field,
    model_validator,
    field_validator,
    NonNegativeFloat,
    HttpUrl,
)


from logdetective.constants import (
    DEFAULT_TEMPERATURE,
    LLM_DEFAULT_MAX_QUEUE_SIZE,
    LLM_DEFAULT_REQUESTS_PER_MINUTE,
    SYSTEM_ROLE_DEFAULT,
    USER_ROLE_DEFAULT,
)
from logdetective.utils import check_csgrep


class BuildLog(BaseModel):
    """Model of data submitted to API."""

    url: str


class JobHook(BaseModel):
    """Model of Job Hook events sent from GitLab.
    Full details of the specification are available at
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The unique job ID on this GitLab instance.
    build_id: int

    # The identifier of the job. We only care about 'build_rpm' and
    # 'build_centos_stream_rpm' jobs.
    build_name: str = Field(pattern=r"^build.*rpm$")

    # A string representing the job status. We only care about 'failed' jobs.
    build_status: str = Field(pattern=r"^failed$")

    # The kind of webhook message. We are only interested in 'build' messages
    # which represents job tasks in a pipeline.
    object_kind: str = Field(pattern=r"^build$")

    # The unique ID of the enclosing pipeline on this GitLab instance.
    pipeline_id: int

    # The unique ID of the project triggering this event
    project_id: int


class EmojiMergeRequest(BaseModel):
    """Model of the 'merge_request' subsection of Emoji webhook messages.
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The identifier of the target project
    target_project_id: int

    # The internal identifier (relative to the target project)
    iid: int


class EmojiHook(BaseModel):
    """Model of Job Hook events sent from GitLab.
    Full details of the specification are available at
    https://docs.gitlab.com/user/project/integrations/webhook_events/#job-events
    This model implements only the fields that we care about. The webhook
    sends many more fields that we will ignore."""

    # The kind of webhook message. We are only interested in 'emoji' messages
    # which represents awarding or revoking emoji reactions on notes.
    object_kind: str = Field(pattern=r"^emoji$")

    # Information about the merge request this emoji applies to, if any.
    merge_request: EmojiMergeRequest = Field(default=None)


class SnippetAnalysis(BaseModel):
    """Model of snippet analysis from LLM."""

    text: str = Field(description="Analysis of log snippet contents.")


class RatedSnippetAnalysis(SnippetAnalysis):
    """Model for rated snippet analysis. This model is used to generate
    json schema for inference with structured output."""

    relevance: int = Field(
        ge=0,
        le=100,
        description="Estimate of likelyhood that snippet contains an error, "
        "with 0 standing for completely unlikely, 100 for absolutely certain.",
    )


class Explanation(BaseModel):
    """Model of snippet or general log explanation from Log Detective"""

    text: str
    logprobs: Optional[List[Dict]] = None

    def __str__(self):
        """Return text of the Explanation"""
        return self.text


class AnalyzedSnippet(BaseModel):
    """Model for snippets already processed by Log Detective.

    explanation: LLM output in form of plain text and logprobs dictionary
    text: original snippet text
    line_number: location of snippet in original log
    """

    explanation: SnippetAnalysis | RatedSnippetAnalysis
    text: str
    line_number: int


class Response(BaseModel):
    """Model of data returned by Log Detective API

    explanation: Explanation
    response_certainty: float
    """

    explanation: Explanation
    response_certainty: float


class StagedResponse(Response):
    """Model of data returned by Log Detective API when called when staged response
    is requested. Contains list of reponses to prompts for individual snippets.

    explanation: Explanation
    response_certainty: float
    snippets: list of AnalyzedSnippet objects
    """

    snippets: List[AnalyzedSnippet]


class KojiStagedResponse(BaseModel):
    """Model of data returned by Log Detective API when called when a Koji build
    analysis is requested. Contains list of reponses to prompts for individual
    snippets.
    """

    task_id: int
    log_file_name: str
    response: StagedResponse


class InferenceConfig(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Model for inference configuration of logdetective server."""

    max_tokens: int = -1
    log_probs: bool = True
    url: str = ""
    # OpenAI client library requires a string to be specified for API token
    # even if it is not checked on the server side
    api_token: str = "None"
    model: str = "default-model"
    temperature: NonNegativeFloat = DEFAULT_TEMPERATURE
    max_queue_size: int = LLM_DEFAULT_MAX_QUEUE_SIZE
    http_timeout: float = 5.0
    user_role: str = USER_ROLE_DEFAULT
    system_role: str = SYSTEM_ROLE_DEFAULT
    llm_api_timeout: float = 15.0
    requests_per_minute: int = LLM_DEFAULT_REQUESTS_PER_MINUTE


class ExtractorConfig(BaseModel):
    """Model for extractor configuration of logdetective server."""

    max_clusters: int = 8
    verbose: bool = False
    max_snippet_len: int = 2000
    csgrep: bool = False

    @field_validator("csgrep", mode="before")
    @classmethod
    def verify_csgrep(cls, v: bool):
        """Verify presence of csgrep binary if csgrep extractor is requested."""
        if v and not check_csgrep():
            raise ValueError(
                "Requested csgrep extractor but `csgrep` binary is not in the PATH"
            )
        return v


class GitLabInstanceConfig(BaseModel):  # pylint: disable=too-many-instance-attributes
    """Model for GitLab configuration of logdetective server."""

    name: str
    url: str = "https://gitlab.com"
    # Path to API of the gitlab instance, assuming `url` as prefix.
    api_path: str = "/api/v4"
    api_token: Optional[str] = None

    # This is a list to support key rotation.
    # When the key is being changed, we will add the new key as a new entry in
    # the configuration and then remove the old key once all of the client
    # webhook configurations have been updated.
    # If this option is left empty or unspecified, all requests will be
    # considered authorized.
    webhook_secrets: Optional[List[str]] = None

    timeout: float = 5.0

    # Maximum size of artifacts.zip in MiB. (default: 300 MiB)
    max_artifact_size: int = 300 * 1024 * 1024

    @field_validator("max_artifact_size", mode="before")
    @classmethod
    def megabytes_to_bytes(cls, v: Any):
        """Convert max_artifact_size from megabytes to bytes."""
        if isinstance(v, int):
            return v * 1024 * 1024
        return 300 * 1024 * 1024


class GitLabConfig(BaseModel):
    """Model for GitLab configuration of logdetective server."""

    instances: Dict[str, GitLabInstanceConfig] = {}

    @model_validator(mode="before")
    @classmethod
    def set_gitlab_instance_configs(cls, data: Any):
        """Initialize configuration for each GitLab instance"""
        if not isinstance(data, dict):
            return data

        instances = {}
        for instance_name, instance_data in data.items():
            instance = GitLabInstanceConfig(name=instance_name, **instance_data)
            instances[instance.url] = instance

        return {"instances": instances}


class KojiInstanceConfig(BaseModel):
    """Model for Koji configuration of logdetective server."""

    name: str = ""
    xmlrpc_url: str = "https://koji.fedoraproject.org/kojihub"
    tokens: List[str] = []


class KojiConfig(BaseModel):
    """Model for Koji configuration of logdetective server."""

    instances: Dict[str, KojiInstanceConfig] = {}
    analysis_timeout: int = 15
    max_artifact_size: int = 300 * 1024 * 1024

    @field_validator("max_artifact_size", mode="before")
    @classmethod
    def megabytes_to_bytes(cls, v: Any):
        """Convert max_artifact_size from megabytes to bytes."""
        if isinstance(v, int):
            return v * 1024 * 1024
        return 300 * 1024 * 1024

    @model_validator(mode="before")
    @classmethod
    def set_koji_instance_configs(cls, data: Any):
        """Initialize configuration for each Koji instance."""
        if isinstance(data, dict):
            instances = {}
            for instance_name, instance_data in data.get("instances", {}).items():
                instances[instance_name] = KojiInstanceConfig(
                    name=instance_name, **instance_data
                )
            data["instances"] = instances
        return data


class LogConfig(BaseModel):
    """Logging configuration"""

    name: str = "logdetective"
    level_stream: str | int = "INFO"
    level_file: str | int = "INFO"
    path: str | None = None
    format: str = "%(levelname)s:%(name)s:%(asctime)s:%(message)s"
    datefmt: str = "%Y-%m-%d %H:%M:%S"


class GeneralConfig(BaseModel):
    """General config options for Log Detective"""

    packages: List[str] = []
    excluded_packages: List[str] = []
    devmode: bool = False
    sentry_dsn: HttpUrl | None = None
    collect_emojis_interval: int = 60 * 60  # seconds
    top_k_snippets: int = 0
    report_certainty: bool = False


class Config(BaseModel):
    """Model for configuration of logdetective server."""

    log: LogConfig = Field(default_factory=LogConfig)
    inference: InferenceConfig = Field(default_factory=InferenceConfig)
    snippet_inference: InferenceConfig = Field(default_factory=InferenceConfig)
    # TODO(jpodivin): Extend to work with multiple extractor configs
    extractor: ExtractorConfig = Field(default_factory=ExtractorConfig)
    gitlab: GitLabConfig = Field(default_factory=GitLabConfig)
    koji: KojiConfig = Field(default_factory=KojiConfig)
    general: GeneralConfig = Field(default_factory=GeneralConfig)

    @model_validator(mode="before")
    @classmethod
    def default_snippet_inference(cls, data: Any):
        """Use base inference configuration, if specific snippet configuration isn't provided."""
        if isinstance(data, dict):
            if "snippet_inference" not in data and "inference" in data:
                data["snippet_inference"] = data["inference"]
        return data


class TimePeriod(BaseModel):
    """Specification for a period of time.

    If no indication is given
    it falls back to a 2 days period of time.

    Can't be smaller than a hour"""

    weeks: Optional[int] = None
    days: Optional[int] = None
    hours: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def check_exclusive_fields(cls, data):
        """Check that only one key between weeks, days and hours is defined,
        if no period is specified, fall back to 2 days."""
        if isinstance(data, dict):
            how_many_fields = sum(
                1
                for field in ["weeks", "days", "hours"]
                if field in data and data[field] is not None
            )

            if how_many_fields == 0:
                data["days"] = 2  # by default fallback to a 2 days period

            if how_many_fields > 1:
                raise ValueError("Only one of months, weeks, days, or hours can be set")

        return data

    @field_validator("weeks", "days", "hours")
    @classmethod
    def check_positive(cls, v):
        """Check that the given value is positive"""
        if v is not None and v <= 0:
            raise ValueError("Time period must be positive")
        return v

    def get_time_period(self) -> datetime.timedelta:
        """Get the period of time represented by this input model.
        Will default to 2 days, if no period is set.

        Returns:
            datetime.timedelta: The time period as a timedelta object.
        """
        delta = None
        if self.weeks:
            delta = datetime.timedelta(weeks=self.weeks)
        elif self.days:
            delta = datetime.timedelta(days=self.days)
        elif self.hours:
            delta = datetime.timedelta(hours=self.hours)
        else:
            delta = datetime.timedelta(days=2)
        return delta

    def get_period_start_time(
        self, end_time: Optional[datetime.datetime] = None
    ) -> datetime.datetime:
        """Calculate the start time of this period based on the end time.

        Args:
            end_time (datetime.datetime, optional): The end time of the period.
                Defaults to current UTC time if not provided.

        Returns:
            datetime.datetime: The start time of the period.
        """
        time = end_time or datetime.datetime.now(datetime.timezone.utc)
        if time.tzinfo is None:
            time = time.replace(tzinfo=datetime.timezone.utc)
        return time - self.get_time_period()


class MetricTimeSeries(BaseModel):
    """Recorded values of given metric"""
    metric: str
    timestamps: List[datetime.datetime]
    values: List[float]


class MetricResponse(BaseModel):
    """Requested metrics"""
    time_series: List[MetricTimeSeries]



from datetime import datetime, timezone
from typing import Optional
from jinja2 import Environment, FileSystemLoader, Template

from logdetective.models import PromptConfig


class PromptManager:  # pylint: disable=too-many-instance-attributes
    """Manages prompts defined as jinja templates"""
    _tmp_env: Environment

    # Templates for system prompts
    _default_system_prompt_template: Template
    _snippet_system_prompt_template: Template
    _staged_system_prompt_template: Template

    # Templates for messages
    default_message_template: Template
    snippet_message_template: Template
    staged_message_template: Template

    _references: Optional[list[dict[str, str]]] = None

    def __init__(
        self, prompts_path: str, prompts_configuration: Optional[PromptConfig] = None
    ) -> None:
        self._tmp_env = Environment(loader=FileSystemLoader(prompts_path))

        self._default_system_prompt_template = self._tmp_env.get_template(
            "system_prompt.j2"
        )
        self._snippet_system_prompt_template = self._tmp_env.get_template(
            "snippet_system_prompt.j2"
        )
        self._staged_system_prompt_template = self._tmp_env.get_template(
            "staged_system_prompt.j2"
        )

        self.default_message_template = self._tmp_env.get_template(
            "message_template.j2"
        )
        self.snippet_message_template = self._tmp_env.get_template(
            "snippet_message_template.j2"
        )
        self.staged_message_template = self._tmp_env.get_template(
            "staged_message_template.j2"
        )

        if prompts_configuration:
            self._references = prompts_configuration.references

    # To maintain backward compatibility with `logdetective.models.PromptConfig`
    @property
    def default_system_prompt(self) -> str:
        """Render system prompt from a template"""
        return self._default_system_prompt_template.render(
            system_time=datetime.now(timezone.utc), references=self._references
        )

    @property
    def snippet_system_prompt(self) -> str:
        """Render system prompt from a template"""
        return self._snippet_system_prompt_template.render(
            system_time=datetime.now(timezone.utc), references=self._references
        )

    @property
    def staged_system_prompt(self) -> str:
        """Render system prompt from a template"""
        return self._staged_system_prompt_template.render(
            system_time=datetime.now(timezone.utc), references=self._references
        )

    @property
    def prompt_template(self) -> str:
        """Render message prompt from the template"""
        return self.default_message_template.render()

    @property
    def snippet_prompt_template(self) -> str:
        """Render message prompt from the template"""
        return self.snippet_message_template.render()

    @property
    def prompt_template_staged(self) -> str:
        """Render message prompt from the template"""
        return self.staged_message_template.render()
